/* vim: set ai et ts=4 sw=4: */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include "file-chooser.h"
#include "file-list-view.h"
#include "file-icon-view.h"
#include "file-tree-view.h"
#include "file-ladder-view.h"
#include "util.h"

/*-----------------------------------------------------------------*/

typedef enum _FileChooserViews {
    ICON_VIEW,
    TREE_VIEW,
    LADDER_VIEW,
    N_VIEWS,
} FileChooserViews;

static const gchar* views_icon[] = {
    "icon-view.png",
    "tree-view.png",
    "ladder-view.png",
    NULL,
};

static void file_chooser_create_file_list_views(FileChooser* filechooser);

/*-----------------------------------------------------------------*/

static void file_list_view_selected(GtkWidget* widget, gpointer data);

/*-----------------------------------------------------------------*/

//
// boiler plate for FileChooser
//

static void file_chooser_class_init(FileChooserClass* klass);
static void file_chooser_init(FileChooser* fs);

static GtkWidgetClass *parent_class = NULL;

GType file_chooser_get_type() 
{ 
    static GType file_chooser_type = 0; 

    if(!file_chooser_type) { 
        static GTypeInfo file_chooser_info = { 
            sizeof(FileChooserClass), 
            NULL,
            NULL,
            (GClassInitFunc)file_chooser_class_init, 
            NULL,
            NULL,
            sizeof(FileChooser), 
            0,
            (GInstanceInitFunc)file_chooser_init
        }; 

        file_chooser_type = g_type_register_static(GTK_TYPE_DIALOG,
                                                   "FileChooser",
                                                   &file_chooser_info,
                                                   0); 
    } 

    return file_chooser_type; 
}


static void file_chooser_class_init(FileChooserClass* klass)
{ 
    // nothing to do
    GObjectClass* g_object_class;
    GtkObjectClass* object_class;
    GtkWidgetClass* widget_class;

    g_object_class = G_OBJECT_CLASS(klass);
    object_class = (GtkObjectClass*)klass;
    widget_class = (GtkWidgetClass*)klass;

    parent_class = g_type_class_peek_parent(klass);

    // install properties
    // setup signals
}


static void file_chooser_init(FileChooser* filechooser)
{ 
    GtkWidget* button;
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* label;
    GtkWidget* table;
    GtkWidget* combo;
    GtkWidget* notebook;

    /* main widgets: location entry, favorate buttons, file list widgets */
    vbox = GTK_DIALOG(filechooser)->vbox;
    gtk_box_set_homogeneous(GTK_BOX(vbox), FALSE);
    gtk_box_set_spacing(GTK_BOX(vbox), 5);

    hbox = gtk_hbox_new(FALSE, 3);
    gtk_box_set_spacing(GTK_BOX(hbox), 5);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);

    // location
    label = gtk_label_new(_("Location"));
    gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

    combo = gtk_combo_new();
    gtk_box_pack_start(GTK_BOX(hbox), combo, FALSE, FALSE, 0);

    // add & remove favorites buttons
    button = gtk_button_new_from_stock("gtk-add");
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);

    button = gtk_button_new_from_stock("gtk-remove");
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);

    // padding
    label = gtk_label_new("");
    gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);

    /* file list view select buttons */
    filechooser->select_view_buttons = gtk_hbox_new(TRUE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), filechooser->select_view_buttons,
                       FALSE, FALSE, 0);

    /* file list widget container */
    notebook = gtk_notebook_new();
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
    gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);
    gtk_box_pack_start(GTK_BOX(vbox), notebook, TRUE, TRUE, 0);
    filechooser->notebook = notebook;

    // filename entry and filter combo
    table = gtk_table_new(4, 2, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, TRUE, 0);

    // filename
    label = gtk_label_new(_("Filename"));
    gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
    gtk_table_attach(GTK_TABLE(table), label,
                     0, 1, 2, 3, GTK_FILL, 0, 0, 0);

    filechooser->filename_ent = gtk_entry_new();
    gtk_table_attach(GTK_TABLE(table), filechooser->filename_ent,
                     1, 2, 2, 3, GTK_EXPAND | GTK_FILL, 0, 0, 0);

    // filter
    label = gtk_label_new(_("Filter"));
    gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
    gtk_table_attach(GTK_TABLE(table), label,
                     0, 1, 3, 4, GTK_FILL, 0, 0, 0);

    combo = gtk_combo_new();
    gtk_table_attach(GTK_TABLE(table), combo,
                     1, 2, 3, 4, GTK_EXPAND | GTK_FILL, 0, 0, 0);

    /* action buttons */
    gtk_dialog_add_button(GTK_DIALOG(filechooser), GTK_STOCK_CANCEL,
                                                   GTK_RESPONSE_CANCEL);
    button = gtk_dialog_add_button(GTK_DIALOG(filechooser),
                                   GTK_STOCK_OK,
                                   GTK_RESPONSE_OK);

    /* create file list view */
    file_chooser_create_file_list_views(filechooser);

    gtk_widget_show_all(vbox);

    gtk_widget_grab_default(button);
}


GtkWidget*
file_chooser_new(void)
{
    FileChooser* filechooser;

    filechooser = g_object_new(file_chooser_get_type(), NULL);

    gtk_window_set_default_size(GTK_WINDOW(filechooser), 400, 400);
    gtk_container_set_border_width (GTK_CONTAINER (filechooser), 10);
    gtk_dialog_set_has_separator(GTK_DIALOG(filechooser), FALSE);

    return GTK_WIDGET(filechooser);
}


void file_chooser_free(FileChooser * fs)
{
    g_free(fs->title);
    g_free(fs->cur_dir);
    g_strfreev(fs->filters);
    g_free(fs->file);
    g_strfreev(fs->files);
}

static void
file_list_view_selected(GtkWidget* widget, gpointer data)
{
    GtkWidget* notebook;
    GtkWidget* current_view;
    gboolean state;
    gint page_num;
    gchar* current_dir;

    state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
    if (state == FALSE)
        return;

    notebook = g_object_get_data(G_OBJECT(widget), "filechooser-notebook");

    page_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
    current_view = gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page_num);
    current_dir = file_list_view_get_current_dir(FILE_LIST_VIEW(current_view));
    file_list_view_set_current_dir(FILE_LIST_VIEW(data), current_dir);
    g_free(current_dir);

    page_num = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), GTK_WIDGET(data));
    gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), page_num);
}

void
file_chooser_append_file_list_view(FileChooser* filechooser,
                                   GtkWidget* widget,
                                   GtkWidget* pixmap)
{
    static GSList* file_list_view_group = NULL;
    GtkWidget* notebook;
    GtkWidget* button;

    notebook = filechooser->notebook;
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook), widget, NULL);

    button = gtk_radio_button_new(file_list_view_group);
    gtk_toggle_button_set_mode (GTK_TOGGLE_BUTTON (button), FALSE);
    g_object_set_data(G_OBJECT(button), "filechooser-notebook", notebook);
    gtk_container_add(GTK_CONTAINER(button), pixmap);
    gtk_box_pack_start(GTK_BOX(filechooser->select_view_buttons),
                       button, FALSE, TRUE, 0);
    g_signal_connect(G_OBJECT(button), "toggled",
                     G_CALLBACK(file_list_view_selected), widget);
    file_list_view_group = gtk_radio_button_get_group(button);

    gtk_widget_show(button);
}


void file_chooser_set_flags(FileChooser* fs, FileChooserFlags flags)
{
    fs->flags = flags;
}


void file_chooser_set_cur_dir(FileChooser* fs, gchar* cur_dir)
{
    g_free(fs->cur_dir);
    fs->cur_dir = g_strdup(cur_dir);
}


void file_chooser_set_cur_filters(FileChooser* fs, gchar** filters)
{
    g_strfreev(fs->filters);
    //fs->filters = g_strdup(filters);
}


gchar* file_chooser_get_file(FileChooser* fs)
{
    return fs->file;
}


gchar** file_chooser_get_files(FileChooser* fs)
{
    return fs->files;
}

static void
file_chooser_create_file_list_views(FileChooser* filechooser)
{
    GtkWidget* pixmap;
    GtkWidget* view;

    /* icon view */
    view = file_icon_view_new();
    pixmap = create_pixmap("icon-view.png");
    file_chooser_append_file_list_view(filechooser, view, pixmap);
    gtk_widget_show_all(view);

    /* tree view */
    view = file_tree_view_new();
    pixmap = create_pixmap("tree-view.png");
    file_chooser_append_file_list_view(filechooser, view, pixmap);
    gtk_widget_show_all(view);

    /* column view */
    view = file_column_view_new();
    pixmap = create_pixmap("ladder-view.png");
    file_chooser_append_file_list_view(filechooser, view, pixmap);
    gtk_widget_show_all(view);
}

