
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include "file-column-view.h"

typedef enum _FileColumnViewColumnIndex {
	ICON_COLUMN,
	NAMEUTF8_COLUMN,
	NAME_COLUMN,
	IS_DIR_COLUMN,
	N_COLUMNS,
} FileColumnViewColumnIndex;

static const gchar* DIR_ICON = "/usr/share/pixmaps/nautilus/default/i-directory-12.png";
static const gchar* FILE_ICON = "/usr/share/pixmaps/nautilus/default/i-regular-12.png";

typedef struct _ColumnItem ColumnItem;

struct _ColumnItem {
    FileColumnView* view;
    GtkWidget* widget;
	GtkTreeModel* model;
    gchar* path;
};


ColumnItem* column_item_new(FileColumnView* view);

static void add_dir(FileColumnView* view, ColumnItem* item,
        const gchar* parent_dir);
static void add_file(FileColumnView* view, ColumnItem* item,
        const gchar* parent_dir, const gchar* name);

static void selection_changed_cb(
        GtkTreeSelection* selection, ColumnItem* view);

static void row_activated_cb(
        GtkTreeView* tv, GtkTreePath* path,
        GtkTreeViewColumn* column, ColumnItem* view);


static void   file_column_view_class_init(FileColumnViewClass* klass);
static void   file_column_view_init(FileColumnView* view);

static gchar* file_column_view_get_current_dir(FileListView* view);
static void   file_column_view_set_current_dir(FileListView* view,
                                               const gchar* path);

static void   column_item_load_dir     (ColumnItem* item,
                                        const gchar* dirname);
static void   column_item_append_entry (ColumnItem* item,
                                        const gchar* dirname,
                                        const gchar* name);


static GtkWidgetClass *parent_class = NULL;

GType file_column_view_get_type() 
{ 
    static GType file_column_view_type = 0; 

    if(!file_column_view_type) { 
        static GTypeInfo file_column_view_info = { 
            sizeof(FileColumnViewClass), 
            NULL,
            NULL,
            (GClassInitFunc)file_column_view_class_init, 
            NULL,
            NULL,
            sizeof(FileColumnView), 
            0,
            (GInstanceInitFunc)file_column_view_init
        }; 

        file_column_view_type = g_type_register_static(TYPE_FILE_LIST_VIEW,
                                                       "FileColumnView",
                                                       &file_column_view_info,
                                                       0); 
    } 

    return file_column_view_type; 
}


static void file_column_view_class_init(FileColumnViewClass* klass)
{ 
    // nothing to do
    GtkObjectClass* object_class;
    GtkWidgetClass* widget_class;
    FileListViewClass* view_class;

    object_class = (GtkObjectClass*)klass;
    widget_class = (GtkWidgetClass*)klass;
    view_class = (FileListViewClass*)klass;

    parent_class = g_type_class_peek_parent(klass);

    view_class->get_current_dir = file_column_view_get_current_dir;
    view_class->set_current_dir = file_column_view_set_current_dir;

    // install properties
}


static void file_column_view_init(FileColumnView* view)
{ 
    GtkWidget* scrolled;
    GtkWidget* viewport;
    GtkWidget* hbox;

    /* default values */
    view->columns = NULL;
    view->viewport = NULL;
    view->hbox = NULL;
    view->show_hidden_files = FALSE;

    /* make widget */
    scrolled = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
                                   GTK_POLICY_ALWAYS,
                                   GTK_POLICY_NEVER);
    gtk_container_add(GTK_CONTAINER(view), scrolled);
    gtk_widget_show(scrolled);

    viewport = gtk_viewport_new(NULL, NULL);
    gtk_container_add(GTK_CONTAINER(scrolled), viewport);
    gtk_widget_show(viewport);
    view->viewport = viewport;

    hbox = gtk_hbox_new(TRUE, 0);
    gtk_container_add(GTK_CONTAINER(viewport), hbox);
    gtk_widget_show(hbox);
    view->hbox = hbox;

    /* load directory information */
    file_column_view_set_current_dir(FILE_LIST_VIEW(view), NULL);
}


GtkWidget* file_column_view_new(void)
{
    FileColumnView* view;

    view = g_object_new(TYPE_FILE_COLUMN_VIEW, NULL);

    return GTK_WIDGET(view);
}


static gchar*
file_column_view_get_current_dir(FileListView* view)
{
    return NULL;
}

static gboolean
file_column_view_scroll_to_end(FileColumnView* view)
{
    GtkAdjustment* hadj;

    /* scroll to right end */
    hadj = gtk_viewport_get_hadjustment(GTK_VIEWPORT(view->viewport));
    gtk_adjustment_set_value(GTK_ADJUSTMENT(hadj),
                             hadj->upper - hadj->page_size);
    gtk_adjustment_value_changed(GTK_ADJUSTMENT(hadj));

    return FALSE;
}

static void
file_column_view_append_column(FileColumnView* view, ColumnItem *item)
{
    GtkAdjustment* hadj;

    gtk_box_pack_start(GTK_BOX(view->hbox), item->widget, FALSE, TRUE, 0);
    view->columns = g_list_prepend(view->columns, item);

    gtk_idle_add(file_column_view_scroll_to_end, view);
}

static void
file_column_view_remove_all_columns(FileColumnView* view)
{
    GtkWidget* widget;
    GList*  list;

    list = view->columns;
    while (list != NULL) {
        widget = ((ColumnItem*)(list->data))->widget;
        gtk_widget_destroy(widget);
        list = list->next;
    }
    g_list_free(list);

    view->columns = NULL;
}

static void
file_column_view_set_current_dir(FileListView* view, const gchar* path)
{
    FileColumnView *column_view;
    ColumnItem* item;
    gchar** ptr;
    gchar** dir_list;
    gchar*  oldpath;
    gchar*  curpath;

    column_view = FILE_COLUMN_VIEW(view);

    if (path == NULL)
        path = g_get_home_dir();

    /* remove all column items */
    file_column_view_remove_all_columns(column_view);

    oldpath = g_strdup("/");
    curpath = NULL;
    dir_list = g_strsplit(path, G_DIR_SEPARATOR_S, 0);
    for (ptr = dir_list; ptr != NULL && *ptr != NULL; ptr++) {
        item = column_item_new(column_view);
        curpath = g_build_filename(oldpath, *ptr, NULL);
        column_item_load_dir(item, curpath);
        file_column_view_append_column(column_view, item);

        g_free(oldpath);
        oldpath = curpath;
        g_print("DIR: %s\n", curpath);
    }
    g_free(oldpath);
    g_strfreev(dir_list);
}

ColumnItem* column_item_new(FileColumnView* view)
{
    ColumnItem* item;
    GtkWidget* scrolled;
    GtkWidget* treeview;
    GtkTreeModel* model;
	GtkCellRenderer* renderer;
	GtkTreeViewColumn* column;
    GtkTreeSelection* selection;

    item = g_new0(ColumnItem, 1);

    scrolled = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
                                   GTK_POLICY_NEVER,
                                   GTK_POLICY_ALWAYS);

	model = GTK_TREE_MODEL(gtk_list_store_new(N_COLUMNS,
                                              GDK_TYPE_PIXBUF,   /* icon */
                                              G_TYPE_STRING,     /* name */
                                              G_TYPE_STRING,     /* path */
                                              G_TYPE_BOOLEAN));  /* is dir */

    treeview = gtk_tree_view_new_with_model(model);
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);
    gtk_container_add(GTK_CONTAINER(scrolled), treeview);

    /* icon column */
	renderer = gtk_cell_renderer_pixbuf_new();
	column = gtk_tree_view_column_new_with_attributes("",
                                                      renderer,
                                                      "pixbuf",
                                                      ICON_COLUMN,
                                                      NULL);
    gtk_tree_view_column_set_fixed_width(column, 20);
	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);

    /* name column */
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(_("Name"),
                                                      renderer,
                                                      "text",
                                                      NAME_COLUMN,
                                                      NULL);
    gtk_tree_view_column_set_min_width(column, 50);
    gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);

    /* connect signals */
    g_signal_connect(G_OBJECT(treeview), "row_activated",
                     G_CALLBACK(row_activated_cb), item);

    /* enable multiple selection */
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
    g_signal_connect(G_OBJECT(selection), "changed",
                     G_CALLBACK(selection_changed_cb), item);

    gtk_widget_show_all(scrolled);

    item->widget = scrolled;
    item->model = model;
    item->view = view;
    item->path = NULL;

    return item;
}

static void
column_item_load_dir(ColumnItem* item, const gchar* dirname)
{
	GDir* dir;
	const gchar* name;

    item->path = g_strdup(dirname);

	dir = g_dir_open(dirname, 0, NULL);
    if(dir == NULL)
        return;

    for (name = g_dir_read_name(dir);
         name != NULL;
         name = g_dir_read_name(dir)) {
		// skip hidden files
        if(!item->view->show_hidden_files && name[0] == '.') continue;

        column_item_append_entry(item, dirname, name);
	}
	g_dir_close(dir);
}


static void
column_item_append_entry(ColumnItem* item,
                        const gchar* dirname,
                        const gchar* name)
{
    GtkListStore* store;
	GtkTreeIter iter;
    GdkPixbuf* icon;
	gchar* name_utf8;
	gchar* path;
    gboolean is_dir;

    store = GTK_LIST_STORE(item->model);

    /* make readable path for gtk widget(UTF-8) */
    name_utf8 = g_filename_to_utf8(name, -1, NULL, NULL, NULL);

    /* make full path */
    path = g_build_filename(dirname, name, NULL);

    is_dir = g_file_test(path, G_FILE_TEST_IS_DIR);
    if (is_dir)
        icon = create_pixbuf(DIR_ICON);
    else
        icon = create_pixbuf(FILE_ICON);

    gtk_list_store_append(store, &iter);
    gtk_list_store_set(store, &iter,
                       ICON_COLUMN, icon,
                       NAMEUTF8_COLUMN, name_utf8,
                       NAME_COLUMN, name,
                       IS_DIR_COLUMN, is_dir,
                       -1);

    g_free(name_utf8);
    g_free(path);
}

//
// callbacks
//

static void selection_changed_cb(GtkTreeSelection* selection,
                                 ColumnItem* item)
{
    GtkTreeModel* model;
    GtkTreeIter iter;
    GtkTreePath* path;
    gchar* filename;
    gboolean is_dir;

    gtk_tree_selection_get_selected(selection, &model, &iter);
/*
    gtk_tree_model_get(model, &iter,
            FILENAME_COLUMN, &filename,
            IS_DIR_COLUMN, &is_dir,
            -1);
            */
}


static void 
row_activated_cb(GtkTreeView* treeview, GtkTreePath* path,
        GtkTreeViewColumn* column, ColumnItem* item)
{
    GtkTreeModel* model;
    GtkTreeIter iter;
    gchar* filename;
    gboolean is_dir;
    gchar* abspath;

    model = gtk_tree_view_get_model(treeview);

    gtk_tree_model_get_iter(model, &iter, path);
    gtk_tree_model_get(model, &iter,
                       NAME_COLUMN, &filename,
                       IS_DIR_COLUMN, &is_dir,
                       -1);

    abspath = g_build_filename(item->path, filename, NULL);
    if (is_dir) {
        ColumnItem* new_item;

        new_item = column_item_new(item->view);
        column_item_load_dir(new_item, abspath);
        file_column_view_append_column(item->view, new_item);
    } else {
        g_print("activate file:%s\n", abspath);
    }
    g_free(abspath);
}

/* vim: set ai et ts=4 sw=4: */
