/* vim: set ai et ts=4 sw=4: */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include "file-icon-view.h"

/*-----------------------------------------------------------------*/

static int DEF_ICON_WIDTH = 48;
static const gchar* DIR_ICON = "/usr/share/pixmaps/nautilus/default/i-directory-24.png";
static const gchar* FILE_ICON = "/usr/share/pixmaps/nautilus/default/i-regular-24.png";

/*-----------------------------------------------------------------*/

static void add_dir(FileIconView* view, const gchar* parent_dir);
static void add_file(FileIconView* view,
        const gchar* parent_dir, const gchar* name);

/*-----------------------------------------------------------------*/

static void il_select_icon_cb(GnomeIconList* il, gint idx, GdkEvent* event, FileIconView* view);

/*-----------------------------------------------------------------*/

//
// boiler plate for FileIconView widget
//

static void file_icon_view_class_init(FileIconViewClass* klass);
static void file_icon_view_init(FileIconView* view);

static GtkWidgetClass *parent_class = NULL;

GType file_icon_view_get_type() 
{ 
    static GType file_icon_view_type = 0; 

    if(!file_icon_view_type) { 
        GTypeInfo file_icon_view_info = { 
            sizeof(FileIconViewClass), 
            NULL,
            NULL,
            (GClassInitFunc)file_icon_view_class_init, 
            NULL,
            NULL,
            sizeof(FileIconView), 
            0,
            (GInstanceInitFunc)file_icon_view_init
        }; 

        file_icon_view_type = g_type_register_static(file_list_view_get_type(), "FileIconView", &file_icon_view_info, 0); 
    } 
    return file_icon_view_type; 
}


static void file_icon_view_class_init(FileIconViewClass* klass)
{ 
    // nothing to do
    GtkObjectClass* object_class;
    GtkWidgetClass* widget_class;

    object_class = (GtkObjectClass*)klass;
    widget_class = (GtkWidgetClass*)klass;

    parent_class = gtk_type_class(file_list_view_get_type());
}


static void file_icon_view_init(FileIconView* view)
{ 
    // nothing to do
}

/*-----------------------------------------------------------------*/

GtkWidget* file_icon_view_new(void)
{
    FileIconView* view;

    view = g_object_new(file_icon_view_get_type(), NULL);

    view->sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_add(GTK_CONTAINER(view), view->sw);

    view->il = gnome_icon_list_new(DEF_ICON_WIDTH, NULL, 0);
    gtk_container_add(GTK_CONTAINER(view->sw), view->il);

    g_signal_connect(G_OBJECT(view->il), "select-icon",
            G_CALLBACK(il_select_icon_cb), view);

    gtk_widget_show_all(GTK_WIDGET(view));

    file_icon_view_refresh(view);

    return GTK_WIDGET(view);
}

/*-----------------------------------------------------------------*/

void file_icon_view_refresh(FileIconView* view)
{
    gnome_icon_list_clear(GNOME_ICON_LIST(view->il));

    add_dir(view, "/");
}

/*-----------------------------------------------------------------*/

static void add_dir(FileIconView* view, const gchar* parent_dir)
{
	GDir* dir;
	const gchar* name;

    g_return_if_fail(IS_FILE_ICON_VIEW(view));

	dir = g_dir_open(parent_dir, 0, NULL);
    if(!dir) return;

	while((name = g_dir_read_name(dir)) != NULL) {
		// skip hidden files
        if(view->parent.show_hidden_files && name[0] == '.') continue;

        add_file(view, parent_dir, name);
	}

	g_dir_close(dir);
}


static void add_file(FileIconView* view, const gchar* parent_dir, const gchar* name)
{
	gchar* name_str;
	gchar* filename;
    gboolean is_dir;
    struct stat st;
    int idx;

    g_return_if_fail(IS_FILE_ICON_VIEW(view));

    // make readable path for gtk widget(UTF-8)
    name_str = g_filename_to_utf8(name, -1, NULL, NULL, NULL);

    // make full path
    filename = g_build_filename(parent_dir, name, NULL);

    if(stat(filename, &st)) return;

    is_dir = S_ISDIR(st.st_mode);

    // show directory only
    if(view->parent.show_dir_only && !is_dir)
        return;

    // add it as a child of parent_iter
    idx = gnome_icon_list_append(
            GNOME_ICON_LIST(view->il),
            (is_dir) ? DIR_ICON : FILE_ICON,
            name_str);
    gnome_icon_list_set_icon_data(
            GNOME_ICON_LIST(view->il),
            idx,
            filename);

    //g_free(name_str);
}

/*-----------------------------------------------------------------*/

static void il_select_icon_cb(GnomeIconList* il, gint idx, GdkEvent* event, FileIconView* view)
{
    gchar* filename;

    filename = gnome_icon_list_get_icon_data(il, idx);

    if(g_file_test(filename, G_FILE_TEST_IS_DIR)) {
        gnome_icon_list_clear(GNOME_ICON_LIST(view->il));
        add_dir(view, filename);
    }
}

/*file-icon-view.c*/
