/* vim: set ai et ts=4 sw=4: */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include "file-ladder-view.h"

/*-----------------------------------------------------------------*/

typedef enum _FileLadderViewColumnIndex {
	ICON_COLUMN,
	NAME_COLUMN,
	FILENAME_COLUMN,
	IS_DIR_COLUMN,
	N_COLUMNS,
} FileLadderViewColumnIndex;

static const gchar* DIR_ICON = "/usr/share/pixmaps/nautilus/default/i-directory-12.png";
static const gchar* FILE_ICON = "/usr/share/pixmaps/nautilus/default/i-regular-12.png";

typedef struct _LadderItem {
    FileLadderView* view;
    GtkWidget* sw;
    GtkWidget* tv;
	GtkTreeModel* model;
	GtkTreeSelection* selection;
} LadderItem;

/*-----------------------------------------------------------------*/

LadderItem* ladder_item_new(FileLadderView* view);

static void add_dir(FileLadderView* view, LadderItem* item,
        const gchar* parent_dir);
static void add_file(FileLadderView* view, LadderItem* item,
        const gchar* parent_dir, const gchar* name);

static void selection_changed_cb(
        GtkTreeSelection* selection, LadderItem* view);

static void row_activated_cb(
        GtkTreeView* tv, GtkTreePath* path,
        GtkTreeViewColumn* column, LadderItem* view);

/*-----------------------------------------------------------------*/

//
// boiler plate for FileLadderView widget
//

static void file_ladder_view_class_init(FileLadderViewClass* klass);
static void file_ladder_view_init(FileLadderView* view);

static GtkWidgetClass *parent_class = NULL;

GType file_ladder_view_get_type() 
{ 
    static GType file_ladder_view_type = 0; 

    if(!file_ladder_view_type) { 
        GTypeInfo file_ladder_view_info = { 
            sizeof(FileLadderViewClass), 
            NULL,
            NULL,
            (GClassInitFunc)file_ladder_view_class_init, 
            NULL,
            NULL,
            sizeof(FileLadderView), 
            0,
            (GInstanceInitFunc)file_ladder_view_init
        }; 

        file_ladder_view_type = g_type_register_static(file_list_view_get_type(), "FileLadderView", &file_ladder_view_info, 0); 
    } 
    return file_ladder_view_type; 
}


static void file_ladder_view_class_init(FileLadderViewClass* klass)
{ 
    // nothing to do
    GtkObjectClass* object_class;
    GtkWidgetClass* widget_class;

    object_class = (GtkObjectClass*)klass;
    widget_class = (GtkWidgetClass*)klass;

    parent_class = gtk_type_class(file_list_view_get_type());
}


static void file_ladder_view_init(FileLadderView* view)
{ 
    // nothing to do
}

/*-----------------------------------------------------------------*/

GtkWidget* file_ladder_view_new(void)
{
    FileLadderView* view;

    view = g_object_new(file_ladder_view_get_type(), NULL);

    view->sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(view->sw),
                GTK_POLICY_ALWAYS, GTK_POLICY_NEVER);
    gtk_container_add(GTK_CONTAINER(view), view->sw);

    view->vp = gtk_viewport_new(NULL, NULL);
    gtk_container_add(GTK_CONTAINER(view->sw), view->vp);

    view->hb = gtk_hbox_new(TRUE, 0);
    gtk_container_add(GTK_CONTAINER(view->vp), view->hb);

    file_ladder_view_refresh(view);

    return GTK_WIDGET(view);
}

/*-----------------------------------------------------------------*/

void file_ladder_view_refresh(FileLadderView* view)
{
    LadderItem* item;
    gchar* cur_dir;
    
    view->ladders = NULL;

    //while() {
        cur_dir = "/";
        item = ladder_item_new(view);
        add_dir(view, item, cur_dir);
        view->ladders = g_list_append(view->ladders, item);
    //}

}

/*-----------------------------------------------------------------*/

LadderItem* ladder_item_new(FileLadderView* view)
{
    LadderItem* item;
    GtkWidget* sw;
    GtkWidget* tv;
    GtkTreeModel* model;
	GtkCellRenderer* renderer;
	GtkTreeViewColumn* column;

    item = g_new0(LadderItem, 1);

    item->view = view;

    item->sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(item->sw),
            GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
    gtk_box_pack_start(GTK_BOX(view->hb), item->sw, FALSE, TRUE, 0);

	item->model = GTK_TREE_MODEL(gtk_list_store_new(N_COLUMNS,
			GDK_TYPE_PIXBUF,//icon
			G_TYPE_STRING,  //name
			G_TYPE_STRING,  //filename
            G_TYPE_BOOLEAN  //is_dir
			));

    item->tv = gtk_tree_view_new_with_model(item->model);
    gtk_container_add(GTK_CONTAINER(item->sw), item->tv);

    // icon column
	renderer = gtk_cell_renderer_pixbuf_new();
	column = gtk_tree_view_column_new_with_attributes(
			"", renderer, "pixbuf", ICON_COLUMN, NULL);
    gtk_tree_view_column_set_fixed_width(column, 20);
	gtk_tree_view_append_column(GTK_TREE_VIEW(item->tv), column);

    // name column
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			_("Name"), renderer, "text", NAME_COLUMN, NULL);
    gtk_tree_view_column_set_min_width(column, 50);
    gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(item->tv), column);

    /*
    // filename column
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			_("Filename"), renderer, "text", FILENAME_COLUMN, NULL);
    gtk_tree_view_column_set_min_width(column, 30);
    gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(item->tv), column);

    // type column
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			_("Directory"), renderer, "text", IS_DIR_COLUMN, NULL);
    gtk_tree_view_column_set_min_width(column, 30);
    gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(item->tv), column);
    */

    // connect signals
    g_signal_connect(G_OBJECT(item->tv), "row_activated",
            G_CALLBACK(row_activated_cb), item);

    // enable multiple selection
    item->selection = gtk_tree_view_get_selection(
            GTK_TREE_VIEW(item->tv));
    g_signal_connect(G_OBJECT(item->selection), "changed",
            G_CALLBACK(selection_changed_cb), item);

    gtk_widget_show_all(item->sw);

    return item;
}


static void add_dir(FileLadderView* view, LadderItem* item,
        const gchar* parent_dir)
{
	GDir* dir;
	const gchar* name;

    g_return_if_fail(IS_FILE_LADDER_VIEW(view));

	dir = g_dir_open(parent_dir, 0, NULL);
    if(!dir) return;

	while((name = g_dir_read_name(dir)) != NULL) {
		// skip hidden files
        if(view->parent.show_hidden_file && name[0] == '.') continue;

        g_print("xxx");
        add_file(view, item, parent_dir, name);
        g_print("yyy");
	}

	g_dir_close(dir);
}


static void add_file(FileLadderView* view, LadderItem* item,
        const gchar* parent_dir, const gchar* name)
{
    GtkListStore* store;
	gchar* name_str;
	gchar* filename;
    gboolean is_dir;
	GtkTreeIter iter;
    struct stat st;

    g_return_if_fail(IS_FILE_LADDER_VIEW(view));

    store = GTK_LIST_STORE(item->model);

    // make readable path for gtk widget(UTF-8)
    name_str = g_filename_to_utf8(name, -1, NULL, NULL, NULL);

    // make full path
    filename = g_build_filename(parent_dir, name, NULL);

    if(stat(filename, &st)) return;

    is_dir = S_ISDIR(st.st_mode);

    // show directory only
    if(view->parent.show_dir_only && !is_dir)
        return;

    // add it as a child of parent_iter
    gtk_list_store_append(store,
            &iter);
    gtk_list_store_set(store, &iter,
            ICON_COLUMN, create_pixbuf((is_dir) ? DIR_ICON : FILE_ICON),
            NAME_COLUMN, name_str,
            FILENAME_COLUMN, filename,
            IS_DIR_COLUMN, is_dir,
            -1);

    g_free(name_str);
    g_free(filename);
}

/*-----------------------------------------------------------------*/

//
// callbacks
//

static void selection_changed_cb(
        GtkTreeSelection* selection, LadderItem* item)
{
    GtkTreeModel* model;
    GtkTreeIter iter;
    GtkTreePath* path;
    gchar* filename;
    gboolean is_dir;

    gtk_tree_selection_get_selected(selection, &model, &iter);
/*
    gtk_tree_model_get(model, &iter,
            FILENAME_COLUMN, &filename,
            IS_DIR_COLUMN, &is_dir,
            -1);
            */
}


static void row_activated_cb(
        GtkTreeView* tv, GtkTreePath* path,
        GtkTreeViewColumn* column, LadderItem* item)
{
    GtkTreeModel* model;
    GtkTreeIter iter;
    gchar* filename;
    gboolean is_dir;

    model = gtk_tree_view_get_model(tv);

    gtk_tree_model_get_iter(model, &iter, path);

    gtk_tree_model_get(model, &iter,
            FILENAME_COLUMN, &filename,
            IS_DIR_COLUMN, &is_dir,
            -1);

    if(is_dir) {
        LadderItem* new_item;
        g_print("activate dir:%s\n", filename);
        g_print("000");
        new_item = ladder_item_new(item->view);
        g_print("111");
        add_dir(item->view, new_item, filename);
        g_print("222");
        item->view->ladders = g_list_append(item->view->ladders, new_item);
        g_print("333");
    } else {
        g_print("activate file:%s\n", filename);
    }
}

/*file-ladder-view.c*/
