/* vim: set ai et ts=4 sw=4: */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include "file-tree-view.h"

/*-----------------------------------------------------------------*/

typedef enum _FileTreeViewColumnIndex {
	ICON_COLUMN,
	NAME_COLUMN,
    TIME_COLUMN,
	SIZE_COLUMN,
    TYPE_COLUMN,
	FILENAME_COLUMN,
	N_COLUMNS,
} FileTreeViewColumnIndex;

static const gchar* DIR_OPEN_ICON = "dir-open.png";
static const gchar* DIR_CLOSE_ICON = "dir-open.png";
static const gchar* FILE_ICON = "file.png";

/*-----------------------------------------------------------------*/

static void add_dir(FileTreeView* view,
        const gchar* parent_dir, GtkTreeIter* parent_iter);
static void add_file(FileTreeView* view,
        const gchar* parent_dir,  GtkTreeIter* parent_iter,
        const gchar* name);
static void set_icon(FileTreeView* view,
        GtkTreeIter* iter, const gchar* filename);
static void tv_row_expanded_cb(GtkTreeView* tv,
        GtkTreeIter* iter, GtkTreePath* path, FileTreeView* view);
static void tv_row_collapsed_cb(GtkTreeView* tv,
        GtkTreeIter* iter, GtkTreePath* path, FileTreeView* view);

/*-----------------------------------------------------------------*/

//
// boiler plate for FileTreeView widget
//


static void file_tree_view_class_init(FileTreeViewClass* klass);
static void file_tree_view_init(FileTreeView* view);

static GtkWidgetClass *parent_class = NULL;

GType file_tree_view_get_type() 
{ 
    static GType file_tree_view_type = 0; 

    if(!file_tree_view_type) { 
        GTypeInfo file_tree_view_info = { 
            sizeof(FileTreeViewClass), 
            NULL,
            NULL,
            (GClassInitFunc)file_tree_view_class_init, 
            NULL,
            NULL,
            sizeof(FileTreeView), 
            0,
            (GInstanceInitFunc)file_tree_view_init
        }; 

        file_tree_view_type = g_type_register_static(file_list_view_get_type(), "FileTreeView", &file_tree_view_info, 0); 
    } 
    return file_tree_view_type; 
}


static void file_tree_view_class_init(FileTreeViewClass* klass)
{ 
    // nothing to do
    GtkObjectClass* object_class;
    GtkWidgetClass* widget_class;

    object_class = (GtkObjectClass*)klass;
    widget_class = (GtkWidgetClass*)klass;

    parent_class = gtk_type_class(file_list_view_get_type());
}


static void file_tree_view_init(FileTreeView* view)
{ 
    // nothing to do
}

/*-----------------------------------------------------------------*/

GtkWidget* file_tree_view_new(void)
{
    FileTreeView* view;
	GtkCellRenderer* renderer;
	GtkTreeViewColumn* column;

    view = g_object_new(file_tree_view_get_type(), NULL);

    view->sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_add(GTK_CONTAINER(view), view->sw);

	view->model = GTK_TREE_MODEL(gtk_tree_store_new(N_COLUMNS,
			GDK_TYPE_PIXBUF,//icon
			G_TYPE_STRING,  //name
			G_TYPE_STRING,  //time
			G_TYPE_STRING,  //size
			G_TYPE_STRING,  //type
			G_TYPE_STRING   //filename
			));

    view->tv = gtk_tree_view_new_with_model(view->model);
    gtk_container_add(GTK_CONTAINER(view->sw), view->tv);

    /*
    // icon column
	renderer = gtk_cell_renderer_pixbuf_new();
	column = gtk_tree_view_column_new_with_attributes(
			"", renderer, "pixbuf", ICON_COLUMN, NULL);
    gtk_tree_view_column_set_fixed_width(column, 20);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view->tv), column);
    */

    // name column
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			_("Name"), renderer, "text", NAME_COLUMN, NULL);
    gtk_tree_view_column_set_min_width(column, 50);
    gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view->tv), column);

    /* */
    // time column
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			_("Modified"), renderer, "text", TIME_COLUMN, NULL);
    gtk_tree_view_column_set_min_width(column, 30);
    gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view->tv), column);

    // size column
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			_("Size"), renderer, "text", SIZE_COLUMN, NULL);
    gtk_tree_view_column_set_min_width(column, 30);
    gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view->tv), column);

    // type column
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			_("Type"), renderer, "text", TYPE_COLUMN, NULL);
    gtk_tree_view_column_set_min_width(column, 30);
    gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view->tv), column);

    /*
    // filename column
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			_("Filename"), renderer, "text", FILENAME_COLUMN, NULL);
    gtk_tree_view_column_set_min_width(column, 30);
    gtk_tree_view_column_set_resizable(column, TRUE);
	gtk_tree_view_append_column(GTK_TREE_VIEW(view->tv), column);
    */

    g_signal_connect(G_OBJECT(view->tv), "row-expanded",
            G_CALLBACK(tv_row_expanded_cb), view);

    g_signal_connect(G_OBJECT(view->tv), "row-collapsed",
            G_CALLBACK(tv_row_collapsed_cb), view);

    file_tree_view_refresh(view);

    gtk_widget_show_all(GTK_WIDGET(view));

    return GTK_WIDGET(view);
}


void file_tree_view_refresh(FileTreeView* view)
{
    GtkTreePath* root_path;
	const gchar* root_dir = "/";

    gtk_tree_store_clear(GTK_TREE_STORE(view->model));

    add_file(view, "", NULL, root_dir);

    root_path = gtk_tree_path_new_first();
    gtk_tree_view_expand_row(GTK_TREE_VIEW(view->tv), root_path, FALSE);
    gtk_tree_path_free(root_path);
}

/*-----------------------------------------------------------------*/

static void add_dir(FileTreeView* view,
        const gchar* parent_dir, GtkTreeIter* parent_iter)
{
	GDir* dir;
	const gchar* name;

    g_return_if_fail(IS_FILE_LIST_VIEW(view));

	dir = g_dir_open(parent_dir, 0, NULL);
    if(!dir) return;

	while((name = g_dir_read_name(dir)) != NULL) {
		// skip hidden files
        if(view->parent.show_hidden_files && name[0] == '.') continue;

        add_file(view, parent_dir, parent_iter, name);
	}

	g_dir_close(dir);
}


static void add_file(FileTreeView* view, const gchar* parent_dir,  GtkTreeIter* parent_iter, const gchar* name)
{
    GtkTreeStore* store;
	gchar* name_str;
	gchar* time_str;
	gchar* size_str;
	gchar* type_str;
	gchar* filename;
	GtkTreeIter iter;
	GtkTreeIter dummy_iter;
    struct stat st;

    g_return_if_fail(IS_FILE_LIST_VIEW(view));

    store = GTK_TREE_STORE(view->model);

    // make readable path for gtk widget(UTF-8)
    name_str = g_filename_to_utf8(name, -1, NULL, NULL, NULL);

    // make full path
    filename = g_build_filename(parent_dir, name, NULL);

    if(stat(filename, &st)) return;

    // show directory only
    if(view->parent.show_dir_only && !S_ISDIR(st.st_mode))
        return;

    time_str = g_strdup(ctime(&st.st_mtime));
    g_strchomp(time_str);
    size_str = gnome_vfs_format_file_size_for_display(st.st_size);
    type_str = gnome_vfs_mime_get_description(
            gnome_vfs_get_mime_type(filename));
            //gnome_vfs_mime_type_from_name_or_default(filename, GNOME_VFS_MIME_TYPE_UNKNOWN));

    // add it as a child of parent_iter
    gtk_tree_store_append(store,
            &iter, parent_iter);
    gtk_tree_store_set(store, &iter,
            NAME_COLUMN, name_str,
            TIME_COLUMN, time_str,
            SIZE_COLUMN, size_str,
            TYPE_COLUMN, type_str,
            FILENAME_COLUMN, filename,
            -1);

    if(S_ISDIR(st.st_mode)) {
        //set_icon(view, &iter, DIR_CLOSE_ICON);
        // add dummy nodes
        gtk_tree_store_append(store,
                &dummy_iter, &iter);
        gtk_tree_store_set(store, &dummy_iter,
                NAME_COLUMN, "...",
                -1);
    } else {
        //set_icon(view, &iter, FILE_ICON);
    }

    g_free(name_str);
    g_free(time_str);
    g_free(size_str);
    //g_free(type_str);
    g_free(filename);
}

static void set_icon(FileTreeView* view, GtkTreeIter* iter, const gchar* filename)
{
    gtk_tree_store_set(GTK_TREE_STORE(view->model),
            iter, ICON_COLUMN, create_pixbuf(filename), -1);
}

/*-----------------------------------------------------------------*/

static void tv_row_expanded_cb(GtkTreeView* tv,
        GtkTreeIter* iter, GtkTreePath* path, FileTreeView* view)
{
    GtkTreeModel* model = view->model;
    GtkTreeIter child_iter;
    gchar* name_str;
    gchar* filename;

    gtk_tree_model_get(model, iter,
            NAME_COLUMN, &name_str, 
            FILENAME_COLUMN, &filename, 
            -1);

    g_print("expand");
    // remove old child nodes
    if(gtk_tree_model_iter_children(model, &child_iter, iter)) {
        do {
            gtk_tree_store_remove(GTK_TREE_STORE(model), &child_iter);
        } while(gtk_tree_model_iter_next(model, &child_iter));
    }

    // scan directory and add nodes
    add_dir(view, filename, iter);

    //set_icon(view, iter, DIR_OPEN_ICON);

    gtk_tree_view_expand_row(tv, path, FALSE);
}


static void tv_row_collapsed_cb(GtkTreeView* tv,
        GtkTreeIter* iter, GtkTreePath* path, FileTreeView* view)
{
    //set_icon(view, iter, DIR_CLOSE_ICON);
}

/*file-tree-view.c*/
