/* vim: set ai et ts=4 sw=4: */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include "util.h"

// FIXME: more elegant way to detect application pixmaps directory
#define PIXMAP_DIR  PACKAGE_DATA_DIR \
                    G_DIR_SEPARATOR_S \
                    "pixmaps" \
                    G_DIR_SEPARATOR_S \
                    PACKAGE \
                    G_DIR_SEPARATOR_S


/* This is an internally used function to find pixmap files. */
gchar* find_pixmap_file(const gchar* filename)
{
	gchar* pathname;
    // absolute path
    if(filename[0] == '/') {
        if(g_file_test(filename, G_FILE_TEST_EXISTS))
            return g_strdup(filename);
    }

    // relative path
    pathname = g_strdup_printf(PIXMAP_DIR "%s", filename);
    if(g_file_test(pathname, G_FILE_TEST_EXISTS)) {
        return pathname;
    }

    g_warning(_("Couldn't find pixmap file: %s"), pathname);
    g_free(pathname);
    return NULL;
}


GtkWidget* create_pixmap(const gchar* filename)
{
    GtkWidget* widget;
    GdkPixbuf* pixbuf = create_pixbuf(filename);

    if (pixbuf)
        widget = gtk_image_new_from_pixbuf(pixbuf);
    else
        widget = gtk_image_new();
    gtk_widget_show(widget);

    return widget;
}


GtkWidget* create_pixmap_with_size(const gchar* filename, gint width, gint height)
{
    GdkPixbuf* pixbuf = create_pixbuf_with_size(filename, width, height);
    return (pixbuf)
            ? gtk_image_new_from_pixbuf(pixbuf)
            : gtk_image_new();
}


GdkPixbuf* create_pixbuf(const gchar* filename)
{
    GdkPixbuf *pixbuf;
    gchar* pathname = NULL;
    GError* error = NULL;

    if(!filename || !filename[0])
        return NULL;

    pathname = find_pixmap_file(filename);
    if(!pathname) return NULL;

    pixbuf = gdk_pixbuf_new_from_file(pathname, &error);
    if(!pixbuf) {
        fprintf(stderr, "Failed to load pixbuf file: %s: %s\n",
                pathname, error->message);
        g_error_free(error);
    }

    g_free(pathname);
    return pixbuf;
}


GdkPixbuf* create_pixbuf_with_size(const gchar* filename, gint width, gint height)
{
    GdkPixbuf* pixbuf = create_pixbuf(filename);
    if(pixbuf) {
        gdk_pixbuf_scale_simple(pixbuf, width, height, GDK_INTERP_BILINEAR);
        return pixbuf;
    }
    return NULL;
}


gboolean str_has_suffix(const gchar* str, const gchar* suffix,
        gboolean case_sensitive)
{
    int len, suffix_len;
    if((len = strlen(str)) < (suffix_len = strlen(suffix))) {
        return FALSE;
    }
    return (case_sensitive)
        ? !strncmp(str + len - suffix_len, suffix, suffix_len)
        : !strncasecmp(str + len - suffix_len, suffix, suffix_len);
}


gint strfindv(const gchar* str, const gchar** str_array,
        gboolean case_sensitive)
{
    int i;
    for(i = 0; str_array[i] != NULL; i++) {
        if(case_sensitive) {
            if(!strcmp(str, str_array[i])) return i;
        } else {
            if(!strcasecmp(str, str_array[i])) return i;
        }
    }
    return -1;
}


/* g_filename_to_utf8 */
gchar* filename_to_utf8(const gchar* str)
{
    return g_filename_to_utf8(str, -1, NULL, NULL, NULL);
}


/* g_filename_from_utf8 */
gchar* utf8_to_filename(const gchar* str)
{
    return g_filename_from_utf8(str, -1, NULL, NULL, NULL);
}


/** g_locale_to_utf8 */
gchar* locale_to_utf8(const gchar* str)
{
    return g_locale_to_utf8(str, -1, NULL, NULL, NULL);
}


/** g_locale_from_utf8 */
gchar* utf8_to_locale(const gchar* str)
{
    return g_locale_from_utf8(str, -1, NULL, NULL, NULL);
}


gchar* uri_to_filename(const GnomeVFSURI* uri)
{
    gchar* uri_str;
    gchar* filename;

    uri_str = gnome_vfs_uri_to_string(uri,
            GNOME_VFS_URI_HIDE_USER_NAME |
            GNOME_VFS_URI_HIDE_PASSWORD |
            GNOME_VFS_URI_HIDE_HOST_NAME |
            GNOME_VFS_URI_HIDE_HOST_PORT |
            GNOME_VFS_URI_HIDE_TOPLEVEL_METHOD);
    filename = gnome_vfs_unescape_string(uri_str, "/");
    g_free(uri_str);
    return filename;
}


gint show_message_box(GtkWindow* parent, GtkMessageType type,
        GtkButtonsType buttons, const gchar* message)
{
    GtkWidget* dlg;
    gint result;
    dlg = gtk_message_dialog_new(parent,
            GTK_DIALOG_DESTROY_WITH_PARENT,
            type, buttons, message);
    result = gtk_dialog_run(GTK_DIALOG(dlg));
    gtk_widget_destroy(dlg);
    return result;
}

/*util.c*/
