/* Copyright (C) 2005 Free Software Foundation, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author: Kang Jeong-Hee <Keizi@mail.co.kr>
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <libgnomevfs/gnome-vfs.h>
#include <eel/eel-gconf-extensions.h>
#include <libnautilus-extension/nautilus-extension-types.h>
#include <libnautilus-extension/nautilus-menu-provider.h>

/*
 * libnautilus-private/nautilus-file-utilities.h Nautilus will provide
 * runtime symbol. 
 */
char *nautilus_get_desktop_directory_uri (void);

static void seed_file_init (void);
static void seed_populate (const gchar *seed_file);
static void seed_pool_free (gpointer data);

#define FOLDER_OF_FORTUNE_TYPE (folder_of_fortune_get_type ())
#define FOLDER_OF_FORTUNE(o)   (G_TYPE_CHECK_INSTANCE_CAST ((o), FOLDER_OF_FORTUNE_TYPE))

typedef struct {
	GObject parent;
} FolderOfFortune;

typedef struct {
	GObjectClass parent_class;
} FolderOfFortuneClass;

static GType this_type = 0;
static GObjectClass *parent_class = NULL;

#define CONF_PREFIX "/apps/folder-of-fortune"
#define SEED_FILE_KEY CONF_PREFIX "/seed_file"
#define OPEN_FOLDER_KEY CONF_PREFIX "/open_folder"

static gboolean open_folder = FALSE;
static gchar *seed_file = NULL;

typedef struct {
	GList *cur;

	GList *seed;
	GList *prefix;
	GList *suffix;
} SeedPool;
static SeedPool seed_pool;

static void
start_element_handler (GMarkupParseContext *context,
		       const gchar         *element_name,
		       const gchar        **attribute_names,
		       const gchar        **attribute_values,
		       gpointer             user_data,
		       GError             **error)
{
	SeedPool *p = (SeedPool *) user_data;

	if (!strcmp (element_name, "seed")) {
		p->cur = p->seed;
	}
	else if (!strcmp (element_name, "prefix")) {
		p->cur = p->prefix;
	}
	else if (!strcmp (element_name, "suffix")) {
		p->cur = p->suffix;
	}
	else {
		p->cur = NULL;
	}
}

static void
end_element_handler (GMarkupParseContext *context,
		     const gchar         *element_name,
		     gpointer             user_data,
		     GError             **error)
{
	SeedPool *p = (SeedPool *) user_data;

	p->cur = NULL;
}

static void
text_handler (GMarkupParseContext *context,
	      const gchar         *text,
	      gsize                text_len,
	      gpointer             user_data,
	      GError             **error)
{
	SeedPool *p = (SeedPool *) user_data;

	if (!p->cur)
		return;
	/*
	 * FIXME 
	 */
	if (text[0] == 0x0a)
		return;

	p->cur = g_list_append (p->cur, (gpointer) g_strdup (text));
}

static GMarkupParseContext *ctxt = NULL;
static GMarkupParser parser = {
	start_element_handler,
	end_element_handler,
	text_handler,
	NULL,
	NULL
};

static GType
folder_of_fortune_get_type (void)
{
	return this_type;
}

static gchar *
get_fortunate_name_internal (GList *pool, gboolean use_pop)
{
	guint len = g_list_length (pool);
	if (len == 1)
		return NULL;
	gchar *name = g_list_nth_data (pool,
				       g_random_int_range (1,
							   g_list_length
							   (pool)));
	gchar *retval = g_strdup (name);
	if (use_pop)
		pool = g_list_remove (pool, name);
	return retval;
}

static gchar *
get_fortunate_name (const guint repeat)
{
	gchar *data, *ret = NULL, *ret0 = NULL;

	ret = get_fortunate_name_internal (seed_pool.seed, TRUE);
	if (!ret) {
		g_list_free (seed_pool.seed);
		seed_pool.seed = NULL;
		return NULL;
	}

	if (repeat >= 1) {
		ret0 = ret;
		data = get_fortunate_name_internal (seed_pool.prefix, FALSE);
		ret = g_strdup_printf ("%s %s", data, ret);
		if (data)
			g_free (data);
		g_free (ret0);
	}
	if (repeat >= 2) {
		ret0 = ret;
		data = get_fortunate_name_internal (seed_pool.suffix, FALSE);
		ret = g_strdup_printf ("%s %s", ret, data);
		if (data)
			g_free (data);
		g_free (ret0);
	}

	return ret;
}

static GnomeVFSURI *
get_fortunate_uri (GnomeVFSURI *parent_uri, gboolean is_local)
{
	guint repeat = 0;
	do {
		seed_file_init ();
		do {
			gchar *new = NULL;
			GnomeVFSURI *uri = NULL;
			new = get_fortunate_name (repeat);
			if (!new)
				break;
			if (is_local) {
				gchar *new_local =
					g_filename_from_utf8 (new, -1,
							      NULL,
							      NULL,
							      NULL);
				uri = gnome_vfs_uri_append_file_name
					(parent_uri, new_local);
				g_free (new_local);
			}
			else {
				uri = gnome_vfs_uri_append_file_name
					(parent_uri, new);
			}
			g_free (new);
			if (!gnome_vfs_uri_exists (uri))
				return uri;
			if (uri)
				gnome_vfs_uri_unref (uri);
			g_print (" %s exists. left %d node\n",
				 gnome_vfs_uri_get_path (uri),
				 g_list_length (seed_pool.seed));

		}
		while (g_list_length (seed_pool.seed));
		repeat += 1;
	}
	while (repeat < 3);
	return NULL;
}

static void
create_folder (NautilusFileInfo *current_folder)
{
	gchar *parent, *new, *new_local = NULL;
	GnomeVFSURI *parent_uri, *uri = NULL;
	GnomeVFSResult res;
	gboolean is_local = FALSE;

	parent = nautilus_file_info_get_uri (current_folder);
	if (eel_uri_is_desktop (parent)) {
		g_free (parent);
		parent = nautilus_get_desktop_directory_uri ();
	}
	parent_uri = gnome_vfs_uri_new (parent);
	is_local = gnome_vfs_uri_is_local (parent_uri);
	uri = get_fortunate_uri (parent_uri, is_local);

	if (uri) {
		res = gnome_vfs_make_directory_for_uri (uri,
							GNOME_VFS_PERM_USER_ALL
							|
							GNOME_VFS_PERM_GROUP_READ
							|
							GNOME_VFS_PERM_GROUP_EXEC
							|
							GNOME_VFS_PERM_OTHER_READ
							|
							GNOME_VFS_PERM_OTHER_EXEC);
		if (res == GNOME_VFS_OK) {
			if (open_folder) {
				new = gnome_vfs_uri_to_string (uri,
							       GNOME_VFS_URI_HIDE_NONE);
				gnome_url_show (new, NULL);
				g_free (new);
			}
		}
		else {
			g_debug ("Error creating folder : %s",
				 gnome_vfs_uri_get_path (uri));
		}
		gnome_vfs_uri_unref (uri);
	}
	else {
		/*
		 * need warning popup? 
		 */
	}
	gnome_vfs_uri_unref (parent_uri);
}

static void
context_menu_activate (NautilusMenuItem *item,
		       NautilusFileInfo *current_folder)
{
	create_folder (current_folder);
}

static gboolean
is_writable (NautilusFileInfo *cur)
{
	gchar *uri;
	GnomeVFSFileInfo *info;
	gboolean writable = FALSE;

	uri = nautilus_file_info_get_uri (cur);
	if (eel_uri_is_desktop (uri)) {
		writable = TRUE;
		goto early_done;
	}
	if (eel_uri_is_in_trash (uri)) {
		writable = FALSE;
		goto early_done;
	}

	info = gnome_vfs_file_info_new ();
	if (gnome_vfs_get_file_info
	    (uri, info,
	     GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS) != GNOME_VFS_OK) {
		writable = FALSE;
		goto done;
	}
	/*
	 * check file permissions 
	 */
	GnomeVFSURI *vfsuri = gnome_vfs_uri_new (uri);
	if (vfsuri && gnome_vfs_uri_is_local (vfsuri)) {
		writable =
			(info->
			 permissions & GNOME_VFS_PERM_ACCESS_WRITABLE) ==
			GNOME_VFS_PERM_ACCESS_WRITABLE;
	}
	else {
		/*
		 * FIXME remote GnomeVFSFileInfo has no access info. 
		 */
		writable = info->permissions & (GNOME_VFS_PERM_USER_WRITE |
						GNOME_VFS_PERM_GROUP_WRITE |
						GNOME_VFS_PERM_OTHER_WRITE);
	}
	if (vfsuri)
		gnome_vfs_uri_unref (vfsuri);
done:
	gnome_vfs_file_info_unref (info);
early_done:
	g_free (uri);

	return writable;
}

static GList *
folder_of_fortune_get_background_items (NautilusMenuProvider *provider,
					GtkWidget *window,
					NautilusFileInfo *current_folder)
{
	GList *items = NULL;
	NautilusMenuItem *item;

	g_return_val_if_fail (is_writable (current_folder), items);

	item = nautilus_menu_item_new ("Folder of fortune::create_any_folder",
				       _("Create any folder"),
				       _("Create a new folder of random name"),
				       NULL);
	g_signal_connect (item, "activate",
			  G_CALLBACK (context_menu_activate), current_folder);
	items = g_list_prepend (items, item);

	return items;
}

static void
folder_of_fortune_class_init (FolderOfFortuneClass *class)
{
	parent_class = g_type_class_peek_parent (class);
}
static void
folder_of_fortune_init (FolderOfFortune *self)
{
}
static void
menu_provider_iface_init (NautilusMenuProviderIface *iface)
{
	iface->get_background_items = folder_of_fortune_get_background_items;
}

void
folder_of_fortune_register_type (GTypeModule *module)
{
	static const GTypeInfo info = {
		sizeof (FolderOfFortuneClass),
		(GBaseInitFunc) NULL,
		(GBaseFinalizeFunc) NULL,
		(GClassInitFunc) folder_of_fortune_class_init,
		NULL,
		NULL,
		sizeof (FolderOfFortune),
		0,
		(GInstanceInitFunc) folder_of_fortune_init
	};
	static const GInterfaceInfo menu_provider_iface_info = {
		(GInterfaceInitFunc) menu_provider_iface_init,
		NULL,
		NULL
	};

	this_type = g_type_module_register_type (module,
						 G_TYPE_OBJECT,
						 "FolderOfFortune", &info, 0);
	g_type_module_add_interface (module,
				     this_type,
				     NAUTILUS_TYPE_MENU_PROVIDER,
				     &menu_provider_iface_info);
}

static void
seed_pool_free (gpointer data)
{
	SeedPool *p = (SeedPool *) data;

	if (p->seed) {
		if (g_list_length (p->seed))
			g_list_foreach (p->seed, (GFunc) g_free, NULL);
		g_list_free (p->seed);
		p->seed = NULL;
	}
	if (p->prefix) {
		if (g_list_length (p->prefix))
			g_list_foreach (p->prefix, (GFunc) g_free, NULL);
		g_list_free (p->prefix);
		p->prefix = NULL;
	}
	if (p->suffix) {
		if (g_list_length (p->suffix))
			g_list_foreach (p->suffix, (GFunc) g_free, NULL);
		g_list_free (p->suffix);
		p->suffix = NULL;
	}

	p->cur = NULL;
}

static void
seed_populate (const gchar *seed_file)
{
	gchar *contents = NULL;
	gsize length;

	if (!seed_file
	    || !g_file_get_contents (seed_file, &contents, &length, NULL)) {
		contents =
			g_strdup (_
				  ("<prefix>absolute</prefix><prefix>totally</prefix><seed>nothing</seed><seed>empty</seed><suffix>default</suffix>"));
		length = strlen (contents);
	}
	if (ctxt) {
		g_markup_parse_context_free (ctxt);
		seed_pool.cur = NULL;
		seed_pool.seed = g_list_alloc ();
		seed_pool.prefix = g_list_alloc ();
		seed_pool.suffix = g_list_alloc ();
	}

	ctxt = g_markup_parse_context_new (&parser, 0,
					   &seed_pool,
					   (GDestroyNotify) seed_pool_free);
	g_markup_parse_context_parse (ctxt, contents, length, NULL);
	g_markup_parse_context_end_parse (ctxt, NULL);
	g_free (contents);
}

static void
seed_file_init (void)
{
	seed_populate (seed_file);
}

static void
seed_file_changed (GConfClient *client,
		   guint cnxn_id,
		   GConfEntry *entry,
		   gpointer user_data)
{
	if (seed_file)
		g_free (seed_file);
	seed_file = eel_gconf_get_string (SEED_FILE_KEY);
}

static void
open_folder_changed (GConfClient *client,
		     guint cnxn_id,
			 GConfEntry *entry,
			 gpointer user_data)
{
	open_folder = eel_gconf_get_boolean (OPEN_FOLDER_KEY);
}

void
nautilus_module_initialize (GTypeModule *module)
{

	folder_of_fortune_register_type (module);

	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");

	seed_pool.cur = NULL;
	seed_pool.seed = g_list_alloc ();
	seed_pool.prefix = g_list_alloc ();
	seed_pool.suffix = g_list_alloc ();
	if (eel_gconf_monitor_add (CONF_PREFIX)) {
		eel_gconf_notification_add (SEED_FILE_KEY,
					    (GConfClientNotifyFunc)
					    seed_file_changed, NULL);
		eel_gconf_notification_add (OPEN_FOLDER_KEY,
					    (GConfClientNotifyFunc)
					    open_folder_changed, NULL);
	}
	seed_file = eel_gconf_get_string (SEED_FILE_KEY);
	open_folder = eel_gconf_get_boolean (OPEN_FOLDER_KEY);
}

void
nautilus_module_shutdown (void)
{

	g_markup_parse_context_free (ctxt);
}

void
nautilus_module_list_types (const GType **types, int *num_types)
{
	static GType type_list[1];

	type_list[0] = FOLDER_OF_FORTUNE_TYPE;
	*types = type_list;
	*num_types = G_N_ELEMENTS (type_list);
}
